#ifndef _MATRIX_H
#define _MATRIX_H

#include <cmath>
#include "Vect3D.h"

/*
*****************************************************************************************
	THIS Matrix Class is optimized for 3D operations.
	ITS USE DEPENDS ON THE USE OF ITS OWN MEMBER FUNCTIONS
		NAMELY THE TRANSFORM FUNC.
			Currently translate and scale dont work right in my system
			Great for rotations though  :)
	IT IS MEANT TO MIMICK A 4x4 MATRIX BUT...
		-multiplication is only done on 3x3 component of actual 3x4 SIGNIFICANT components
		-Hence operator= only copies first 12 elements
		-HOWEVER a 4x4 structure is still used to provide better alignment in memory
	IT IS NOT SUITED FOR PROJECTION MATRICIES SINCE 4th ROW IS EFFECTIVELY IGNORED
	ALL Matrix ops that depend on that 4th row should not be attempted
	NEEDS FIXME's ALL FFFREAKEEEN OVER THE PLACE   :(
*/

const double _IDENTITY_[16] = {1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1};

class Matrix {
protected:
	double *elements;
public:
	Matrix();
	Matrix(double *);
	Matrix(const Matrix &);
	~Matrix() {delete[] elements;}

	void  set(double *elems) {memcpy(elements,elems,16*sizeof(double));}
	void  setIdentity()			 {memcpy(elements,_IDENTITY_,16*sizeof(double));}
	double operator[](int index) const {return elements[index];}

	Matrix operator* (const Matrix &) const;

	void rotate		(int ,double);
	void rotate		(const Vect3D &);
	void scale		(const Vect3D &);
	void translate(const Vect3D &);

	Matrix transform	(const Vect3D &, const Vect3D &, const Vect3D &) ;
	Vect3D cross			(const Vect3D &) const;
	Vect3D back_rotate(const Vect3D &) const;

	Matrix& operator=(const Matrix &);
};

Matrix::Matrix() {
	elements = new double[16];
	memcpy(elements,_IDENTITY_,16*sizeof(double));
}
Matrix::Matrix(double *elems) {
	elements = new double[16];
	memcpy(elements,elems,16*sizeof(double));
}
Matrix::Matrix(const Matrix &rMat1) {
	elements = new double[16];
	memcpy(elements,rMat1.elements,16*sizeof(double));
}

Matrix Matrix::operator*(const Matrix &rhs) const {
	Matrix temp;
	temp.elements[0]=elements[0]*rhs.elements[0] + elements[1]*rhs.elements[4] +
				elements[2]*rhs[8];
	temp.elements[1]=elements[0]*rhs.elements[1] + elements[1]*rhs.elements[5] +
				elements[2]*rhs[9];
	temp.elements[2]=elements[0]*rhs.elements[2] + elements[1]*rhs.elements[6] +
				elements[2]*rhs[10];

	temp.elements[4]=elements[4]*rhs.elements[0] + elements[5]*rhs.elements[4] +
				elements[6]*rhs[8];
	temp.elements[5]=elements[4]*rhs.elements[1] + elements[5]*rhs.elements[5] +
				elements[6]*rhs[9];
	temp.elements[6]=elements[4]*rhs.elements[2] + elements[5]*rhs.elements[6] +
				elements[6]*rhs[10];

	temp.elements[8]=elements[8]*rhs.elements[0] + elements[9]*rhs.elements[4] +
				elements[10]*rhs[8];
	temp.elements[9]=elements[8]*rhs.elements[1] + elements[9]*rhs.elements[5] +
				elements[10]*rhs[9];
	temp.elements[10]=elements[8]*rhs.elements[2] + elements[9]*rhs.elements[6] +
				elements[10]*rhs[10];

	return temp;
}

void Matrix::rotate(int axis, double angle) {
	Matrix temp;
	double c,s;
	int i = (axis%3) + 1;
	int j = i%3;
	i -= 1;
	c = cos(angle*PI_180);
	s = sin(angle*PI_180);
	temp.elements[(i<<2)+i] = c; temp.elements[(i<<2)+j] = s;
	temp.elements[(j<<2)+j] = c; temp.elements[(j<<2)+i] = -s;
	(*this) = (*this)*temp;
}
void Matrix::rotate(const Vect3D &ro) {
	rotate(1,ro.x);
	rotate(2,ro.y);
	rotate(3,ro.z);
}
void Matrix::scale(const Vect3D &sc) {
	elements[0] = sc.x;
	elements[5] = sc.y;
	elements[10] = sc.z;
}
void Matrix::translate(const Vect3D &tr) {
	elements[3] = -tr.x;
	elements[7] = -tr.y;
	elements[11] = -tr.z;
}

Matrix Matrix::transform(const Vect3D &sc,const Vect3D &tr,const Vect3D &ro) {
	scale(sc);
	rotate(ro);
	translate(tr);
	return *this;
}

Vect3D Matrix::cross(const Vect3D &v) const {
	double x = elements[0]*v.x + elements[1]*v.y + elements[2]*v.z; // + elements[3];
	double y = elements[4]*v.x + elements[5]*v.y + elements[6]*v.z; // + elements[7];
	double z = elements[8]*v.x + elements[9]*v.y + elements[10]*v.z; // + elements[11];
	return Vect3D(x,y,z);
}

Vect3D Matrix::back_rotate(const Vect3D &v) const {
	double x = v.x*elements[0] + v.y*elements[4] + v.z*elements[8];
	double y = v.x*elements[1] + v.y*elements[5] + v.z*elements[9];
	double z = v.x*elements[2] + v.y*elements[6] + v.z*elements[10];
	return Vect3D(x,y,z);
}

Matrix& Matrix::operator=(const Matrix &rhs) {
	if(&rhs != this) {
		memcpy(elements,rhs.elements,12*sizeof(double));
	}
	return *this;
}

#endif